import click

from switchbladecli.cli.bundle_cache import Bundle
from switchbladecli.cli.config import get_switchblade_config


@click.command()
@click.pass_context
def update(ctx):
    project_dir = ctx.obj["project_dir"]
    config_file = ctx.obj["config"]
    verbose = ctx.obj["verbose"]
    update_command(verbose, project_dir, config_file)


def update_command(verbose: bool, project_dir: str, config_file: str):
    config = get_switchblade_config(verbose, project_dir, config_file)
    # Instantiating a Bundle object will fetch the latest bundle from the source or cache
    latest_bundle = Bundle(config)



import click
import os
import shutil

from datetime import datetime, timezone
from github import Github
from pathlib import Path
from tomlkit import dumps, loads

from switchbladecli.utils import hash_dir

BUNDLE_CONFIG_FILES = ["bundle.toml"]
CACHE_FOLDERNAME = ".switchblade-cache"
LATEST_CONFIG_FILENAME = "latest.toml"
UPDATE_LOG_FILENAME = "update.log"


class Bundle:
    """A bundle of dev tools and configs.
    """

    def __init__(self, switchblade_config: dict, version: str = None):
        """Initialize a Bundle object.

        If version is None, the latest bundle will be used (conditionally fetched and cached).

        If a version is passed:
            1. Tries to fetch a bundle with that version from the cache.
            2. If the bundle doesn't exist, an exception is raised.

        Args:
            switchblade_config (dict): The switchblade config.
            version (str, optional): The version of the bundle to fetch. Defaults to None.
        """
        self._switchblade_config = switchblade_config
        if version is None:
            cache = BundleCache(self._switchblade_config)
            bundle = self._fetch_latest(cache)
            self.version = bundle.version
            self.bundle_folder = bundle.bundle_folder
            self.bundle_config = bundle.bundle_config
            cache.set_latest_config(bundle)

        else:
            self.version = version
            self.bundle_folder = self._bundle_folder_from_version(version)
            self.bundle_config = self._load_bundle_config()
    

    def _load_bundle_config(self) -> dict:
        """Get the bundle config from the config file."""
        bundle_config_file = None
        for bundle_config_file_name in BUNDLE_CONFIG_FILES:
            bundle_config_file = self.bundle_folder / bundle_config_file_name
            if bundle_config_file.exists():
                break
        if bundle_config_file is None:
            raise FileNotFoundError(f"No bundle config file found in bundle {self.version}")
        return loads(bundle_config_file.read_text(encoding="utf-8"))
    
    
    def _bundle_folder_from_version(self, version) -> Path:
        """Get the bundle folder."""
        return Path(CACHE_FOLDERNAME) / version

    def get_files(self) -> list:
        """Get the list of files (excluding the config) in the bundle.
        
        Returns a list of file paths relative to the bundle folder.
        """
        return [str(file.relative_to(self.bundle_folder)) for file in self.bundle_folder.rglob("*") if file.name not in BUNDLE_CONFIG_FILES]


    def _fetch_latest(self, cache: "BundleCache") -> "Bundle":
        """Conditionally fetch and return the latest bundle.
        """
        bundle_source_uri = self._switchblade_config["switchblade"]["bundle"]
        click.echo("⚔️ Switchblade checking for updates...")

        # Check the remote version to see if we need to update
        remote_version = None
        pygithub = None
        if bundle_source_uri.startswith("gh:"):
            # Init the Github API
            pygithub = Github(os.environ.get("GITHUB_TOKEN"))
            # Get the org name, repo name and folder(s)
            org_name, repo_name, *folder = bundle_source_uri[3:].split("/")
            # Get the commit SHA from the repo (default branch)
            repo = pygithub.get_repo(f"{org_name}/{repo_name}")
            remote_version = repo.get_branch(repo.default_branch).commit.sha
        else:
            # TODO: Hash/Checksum the local folder
            # If the folder doesn't exist, raise an exception
            bundle_source_folder = Path(bundle_source_uri)
            if not bundle_source_folder.exists():
                raise FileNotFoundError(f"Bundle folder {bundle_source_uri} does not exist.")
            remote_version = hash_dir(bundle_source_uri)

        cache.log(f"UPDATING {remote_version}")

        # Check if the latest version from the remote is already in the local cache
        latest_config = cache.get_latest_config()
        if latest_config is not None and latest_config["commit_sha"] == remote_version:
            # If the latest cached bundle is up to date, just return it
            click.echo(f"⚔️ Switchblade cache already up to date at version {remote_version}")
            cache.log(f"UPDATING {remote_version} NO_CHANGE")
            return Bundle(self._switchblade_config, latest_config["commit_sha"])
        
        # Check if the latest version from the remote is already sitting in an old folder
        if cache.has_bundle(remote_version):
            # If the latest cached bundle is up to date, just return it
            click.echo(f"⚔️ Switchblade cache already up to date at version {remote_version}")
            cache.log(f"UPDATING {remote_version} USING_CACHED")
            return Bundle(self._switchblade_config, remote_version)

        # If the cache is stale, update it
        click.echo(f"⚔️ Switchblade updating tool bundle to version {remote_version}...")
        bundle_folder = self._bundle_folder_from_version(remote_version)
        if bundle_source_uri.startswith("gh:"):
            # Create the bundle folder
            bundle_folder.mkdir(parents=True)
            # Get the contents of the repo
            org_name, repo_name, *folder = bundle_source_uri[3:].split("/")
            repo = pygithub.get_repo(f"{org_name}/{repo_name}")
            repo_contents = repo.get_contents("/".join(folder), ref=remote_version)
            # If the contents are a folder, download it
            if repo_contents.type == "dir":
                for repo_asset in repo_contents:
                    with open(bundle_folder / repo_asset.name, "wb") as file:
                        file.write(repo_asset.decoded_content)
            else:
                # TODO: Use Click exception/error instead
                cache.log(f"UPDATING {remote_version} FAILED")
                raise Exception(
                    f"Bundle URI {bundle_source_uri} is not a folder. Please use a folder in a Github repo."
                )
        else:
            # If the source is a local folder, copy it to the cache
            shutil.copytree(bundle_source_uri, bundle_folder)

        cache.log(f"UPDATING {remote_version} SUCCEEDED")

        # Finally, return the cached bundle
        return Bundle(self._switchblade_config, remote_version)


class BundleCache:

    def __init__(self, switchblade_config: dict):
        self._switchblade_config = switchblade_config
        self._project_folder = Path(self._switchblade_config.get("project_dir"))
        self._cache_folder = self._initialize_cache_folder(self._project_folder)


    def _initialize_cache_folder(self, project_folder: Path):
        """Initialize a new cache folder if it doesn't exist yet.

        Creates:
            - The cache folder
            - A .gitignore file in the cache folder with an asterisk in it
            - An update.log file in the cache folder
        """
        cache_folder = project_folder / CACHE_FOLDERNAME
        if not cache_folder.exists():
            # Make sure the cache folder exists
            cache_folder.mkdir()
            # Add a .gitignore file with an asterisk in it to ensure that the
            # cache folder is never committed to git
            with open(cache_folder / ".gitignore", "w") as gitignore:
                gitignore.write("*")

        # Make sure the update.log file exists
        update_log = cache_folder / UPDATE_LOG_FILENAME
        if not update_log.exists():
            update_log.touch()
        
        return cache_folder


    def get_bundle(self, version: str) -> Bundle:
        """Get a bundle from the cache."""
        return Bundle(self._switchblade_config, version)


    def get_bundles(self) -> list:
        """Get all bundles in the cache."""
        bundles = []
        for bundle_folder in os.listdir(self._cache_folder):
            # Skip non-folders
            if not (self._cache_folder / bundle_folder).is_dir():
                continue
            bundles.append(Bundle(self._switchblade_config, bundle_folder))
        return bundles


    def _get_latest_config_file(self, check_exists: bool = True) -> Path:
        """Get the latest config file."""
        latest_config_file = self._cache_folder / LATEST_CONFIG_FILENAME
        if check_exists and not latest_config_file.exists():
            return None
        return latest_config_file


    def get_latest_config(self) -> dict:
        """Get the latest config."""
        latest_config_file = self._get_latest_config_file()
        # If the latest.toml file doesn't exist, return None
        if latest_config_file is None:
            return None
        latest = loads(latest_config_file.read_text(encoding="utf-8"))
        return latest["latest"]


    def set_latest_config(self, bundle: Bundle):
        """Set the latest config."""
        latest_config_file = self._get_latest_config_file(False)
        with open(latest_config_file, "w") as latest_toml:
            latest_toml.write(
                dumps(
                    {
                        "latest": {
                            "commit_sha": bundle.version,
                            "timestamp": datetime.now(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                            "files": bundle.get_files(),
                        }
                    }
                )
            )


    def get_latest_version(self) -> str:
        """Get the latest version."""
        latest_config = self.get_latest_config()
        if latest_config is None:
            return None
        return latest_config["commit_sha"]


    def get_latest_bundle(self) -> Bundle:
        """Get the latest bundle."""
        latest_config = self.get_latest_config()
        if latest_config is None:
            return None
        return self.get_bundle(latest_config["commit_sha"])


    def log(self, message):
        current_time = datetime.now(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        with open(self._cache_folder / UPDATE_LOG_FILENAME, "a") as update_log:
            update_log.write(f"[{current_time}] {message}\n")


    def has_bundle(self, version: str) -> bool:
        """Check if a version is in the cache."""
        return (self._cache_folder / version).exists()


    def cleanup(self):
        """Remove all bundles from cache except the latest.
        """
        latest_version = self.get_latest_version()
        bundles = self.get_bundles()
        for bundle in bundles:
            if bundle.version != latest_version:
                shutil.rmtree(bundle.bundle_folder)



import hashlib
import os

def sha1OfFile(filepath, use_sha):
    with open(filepath, 'rb') as f:
        while True:
            block = f.read(2**10) # Magic number: one-megabyte blocks.
            if not block: break
            use_sha.update(block)
        return use_sha.hexdigest()

def hash_dir(dir_path, use_sha=None):
    for path, dirs, files in os.walk(dir_path):
        sha = use_sha or hashlib.sha1()
        for file in sorted(files): # we sort to guarantee that files will always go in the same order
            sha1OfFile(os.path.join(path, file), sha)
        for dir in sorted(dirs): # we sort to guarantee that dirs will always go in the same order
            hash_dir(os.path.join(path, dir), sha)
        break # we only need one iteration - to get files and dirs in current directory
    return sha.hexdigest()



import os
from pathlib import Path

import click
from tomlkit import loads


def get_switchblade_config(verbose, project_dir: str, config_file: str):
    config_file = os.path.join(project_dir, config_file)
    if verbose:
        click.echo(f"Loading configuration from {config_file}")

    config = loads(Path(config_file).read_text(encoding="utf-8"))

    config["project_dir"] = project_dir
    config["config_file"] = config_file

    return config



import os

import click

from switchbladecli.cli.update import update
from switchbladecli.cli.lint import lint
from switchbladecli.cli.test import test

ALIASES = {}

CONFIG_FILE_NAMES = [".switchblade", "switchblade.toml"]


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


# We pass the project directory to all subcommands via the context
# so they can use it to find the switchblade.toml file
@click.group(cls=AliasedGroup)
@click.option("--config", help="Path to Switchblade configuration file.")
@click.option("--verbose", is_flag=True, default=False, help="Show verbose output")
@click.pass_context
def switchbladecli(
    ctx, config, verbose
): 
    """Sentient Switchblade

    A part of the Beth Developer Toolbelt
    """
    # Locate the correct Kegstand configuration file
    if config is None:
        for name in CONFIG_FILE_NAMES:
            if os.path.exists(name):
                config = name
                break

    if not os.path.exists(config):
        raise click.ClickException(f"Configuration file not found: {config}")

    project_dir = os.path.abspath(os.path.dirname(config))
    ctx.obj = {
        "config": os.path.abspath(config),
        "project_dir": project_dir,
        "verbose": verbose,
    }


for command in [update, lint, test]:
    switchbladecli.add_command(command)


if __name__ == "__main__":
    switchbladecli(obj={})  # pylint: disable=no-value-for-parameter



from pathlib import Path
import click
import shutil
import subprocess  # nosec

from tomlkit import dumps, loads

from switchbladecli.cli.bundle_cache import Bundle, BundleCache


class PythonPoetry:

    RESET_LOCKFILE_ON_CLEANUP = False
    mode = "python-poetry"

    def __init__(self, config: dict, verbose: bool):
        self._config = config
        self._project_folder = Path(config["project_dir"])
        self._verbose = verbose
        self._cache = BundleCache(self._config)


    # TODO: Make this an override
    def install_dev_tools(self, bundle: Bundle):
        # Read pyproject.toml and remove the sections [tool.poetry.group.switchblade]
        # and [tool.poetry.group.switchblade.dependencies], and add the ones from the
        # bundle.toml file in the commit SHA folder
        self._pyproject_file = self._project_folder / "pyproject.toml"
        self._poetry_lockfile = self._project_folder / "poetry.lock"

        self._pyproject_file_raw_before = self._pyproject_file.read_text(encoding="utf-8")
        self._poetry_lockfile_raw_before = self._poetry_lockfile.read_text(encoding="utf-8")  # TODO: handle missing file
        pyproject_config = loads(self._pyproject_file_raw_before)
        pyproject_config["tool"]["poetry"]["group"][
            "switchblade"
        ] = bundle.bundle_config["tool"]["poetry"]["group"]["switchblade"]
        # pyproject_config["tool"]["poetry"]["group"]["switchblade"][
        #     "dependencies"
        # ] = bundle.bundle_config["tool"]["poetry"]["group"]["switchblade"][
        #     "dependencies"
        # ]
        # Also take all other sections beginning with 'tool.' and add them to the pyproject.toml
        # unless there is already a section with the same name in the pyproject.toml
        for section in bundle.bundle_config["tool"]:
            if not section.startswith("poetry") and section not in pyproject_config["tool"]:
                pyproject_config["tool"][section] = bundle.bundle_config["tool"][
                    section
                ]
        with open(self._pyproject_file, "w") as pyproject_toml:
            pyproject_toml.write(dumps(pyproject_config))

        lock_check = subprocess.run(
            ["poetry", "lock", "--check"],
            cwd=self._project_folder,
            check=False,
            capture_output=True,
        )

        # If the lock file is up to date, we can optimistically skip the update
        if lock_check.returncode == 0:
            print("🔒 poetry.lock is up to date, install skipped")
            return

        subprocess.run(
            ["poetry", "update", "--only", "switchblade"],
            cwd=self._project_folder,
            check=True,
            capture_output=False,
        )


    # TODO: Place this in a superclass
    def copy_config_files(self, bundle: Bundle):
        # For each file in the bundle, check if the file exists in the project folder
        pushed_files = []
        for file in bundle.get_files():
            if (self._project_folder / file).exists():
                # If the file exists, skip it
                self._cache.log(f"SKIPPED {file}")
                continue
            else:
                # If the file doesn't exist, copy it from the commit SHA folder to the project folder
                shutil.copyfile(
                    bundle.bundle_folder / file,
                    self._project_folder / file
                )
                pushed_files.append(file)
                self._cache.log(f"PUSHED {file}")
        return pushed_files


    # TODO: Place this in a superclass
    def run_linters(self, bundle: Bundle):
        succeeded = True
        for linter in bundle.bundle_config["linters"]["all"]:
            print(f"⚔️ Switchblade running {linter}...")
            succeeded = succeeded and self.run_linter(bundle, bundle.bundle_config["linters"][linter])
            self._cache.log(f"RAN {linter}")
        return succeeded


    # TODO: Place this in a superclass
    def run_tests(self, bundle: Bundle):
        succeeded = True
        for test in bundle.bundle_config["tests"]["all"]:
            print(f"⚔️ Switchblade running {test}...")
            succeeded = succeeded and self.run_test(bundle, bundle.bundle_config["tests"][test])
            self._cache.log(f"RAN {test}")
        return succeeded


    # TODO: Make this an override (must return False if the linter fails)
    def run_linter(self, bundle: Bundle, linter: dict):
        linter_command_list = linter["command"].split(" ")
        cmd_result = subprocess.run(
            ["poetry", "run", *linter_command_list], cwd=self._project_folder, check=False
        )
        return cmd_result.returncode == 0


    # TODO: Make this an override (must return False if the test fails)
    def run_test(self, bundle: Bundle, test: dict):
        test_command_list = test["command"].split(" ")
        cmd_result = subprocess.run(
            ["poetry", "run", *test_command_list], cwd=self._project_folder, check=False
        )
        return cmd_result.returncode == 0


    # TODO: Make this an override
    def post_cleanup(self, bundle: Bundle, pushed_files: list):
        # Restore the original pyproject.toml
        if self._pyproject_file_raw_before:
            with open(self._pyproject_file, "w") as pyproject_toml:
                pyproject_toml.write(self._pyproject_file_raw_before)
        # Restore the original poetry.lock
        if self.RESET_LOCKFILE_ON_CLEANUP and self._poetry_lockfile_raw_before:
            with open(self._poetry_lockfile, "w") as poetry_lock:
                poetry_lock.write(self._poetry_lockfile_raw_before)


    # TODO: Place this in a superclass
    def lint(self):
        # Instantiating a Bundle object will fetch the latest bundle from the source or cache
        latest_bundle = Bundle(self._config)

        bundle_config = latest_bundle.bundle_config
        bundle_mode = bundle_config["bundle"]["mode"]

        if bundle_mode != self.mode:
            click.echo(
                f"⚔️ The Switchblade bundle is for a project of mode {bundle_mode}, but this project is of mode {self.mode}.",
                err=True
            )
            raise click.Abort()

        self._cache.log(f"LINTING {latest_bundle.version}")

        pushed_files = []
        success = False
        try:
            print("⚔️ Switchblade installing dev tools...")
            self.install_dev_tools(latest_bundle)

            print("⚔️ Switchblade copying configuration...")
            pushed_files = self.copy_config_files(latest_bundle)

            success = self.run_linters(latest_bundle)

        except Exception:
            self._cache.log(f"LINTING {latest_bundle.version} FAILED_WITH_EXCEPTION")
            raise click.ClickException("Exception detected while running linters, aborting...")

        finally:
            print("⚔️ Switchblade cleaning up...")
            # Remove the files from the project folder that were added in Step 3
            for file in pushed_files:
                (self._project_folder / file).unlink()
                self._cache.log(f"POPPED {file}")

            self.post_cleanup(latest_bundle, pushed_files)

        if success:
            self._cache.log(f"LINTING {latest_bundle.version} FINISHED_SUCCESSFULLY")
            click.echo("⚔️ Linting finished successfully. 😎")
        else:
            self._cache.log(f"LINTING {latest_bundle.version} FINISHED_WITH_ERRORS")
            click.secho("⚔️ Linting finished with errors. 😭", err=True, fg="red", bold=True)


    # TODO: Place this in a superclass
    def test(self):
        # Instantiating a Bundle object will fetch the latest bundle from the source or cache
        latest_bundle = Bundle(self._config)

        bundle_config = latest_bundle.bundle_config
        bundle_mode = bundle_config["bundle"]["mode"]

        if bundle_mode != self.mode:
            click.echo(
                f"⚔️ The Switchblade bundle is for a project of mode {bundle_mode}, but this project is of mode {self.mode}.",
                err=True
            )
            raise click.Abort()

        self._cache.log(f"TESTING {latest_bundle.version}")

        pushed_files = []
        success = False
        try:
            print("⚔️ Switchblade installing dev tools...")
            self.install_dev_tools(latest_bundle)

            print("⚔️ Switchblade copying configuration...")
            pushed_files = self.copy_config_files(latest_bundle)

            success = self.run_tests(latest_bundle)

        except Exception:
            self._cache.log(f"TESTING {latest_bundle.version} FAILED_WITH_EXCEPTION")
            raise click.ClickException("Exception detected while running tests, aborting...")

        finally:
            print("⚔️ Switchblade cleaning up...")
            # Remove the files from the project folder that were added in Step 3
            for file in pushed_files:
                (self._project_folder / file).unlink()
                self._cache.log(f"POPPED {file}")

            self.post_cleanup(latest_bundle, pushed_files)

        if success:
            self._cache.log(f"TESTING {latest_bundle.version} FINISHED_SUCCESSFULLY")
            click.echo("⚔️ Tests finished successfully. 😎")
        else:
            self._cache.log(f"TESTING {latest_bundle.version} FINISHED_WITH_ERRORS")
            click.secho("⚔️ Tests finished with errors. 😭", err=True, fg="red", bold=True)



import click

from switchbladecli.modes.python_poetry import PythonPoetry
from switchbladecli.cli.config import get_switchblade_config


@click.command()
@click.pass_context
def lint(ctx):
    project_dir = ctx.obj["project_dir"]
    config_file = ctx.obj["config"]
    verbose = ctx.obj["verbose"]
    lint_command(verbose, project_dir, config_file)


def lint_command(verbose: bool, project_dir: str, config_file: str):
    config = get_switchblade_config(verbose, project_dir, config_file)

    mode = config["switchblade"]["mode"]

    tool_runner = None
    if mode == "python-poetry":
        tool_runner = PythonPoetry(config, verbose)
    else:
        raise click.ClickException(
            f"Project mode {mode} is not supported by Switchblade."
        )

    tool_runner.lint()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['switchbladecli', 'switchbladecli.cli', 'switchbladecli.modes']

package_data = \
{'': ['*']}

install_requires = \
['PyGithub>=1.58.1,<2.0.0',
 'checksumdir>=1.2.0,<2.0.0',
 'click>=8.0.3,<9.0.0',
 'tomlkit>=0.11.7,<0.12.0']

entry_points = \
{'console_scripts': ['swb = switchbladecli.cli.__main__:switchbladecli']}

setup_kwargs = {
    'name': 'sentient-switchblade',
    'version': '0.1.0',
    'description': 'Unleash Dev Tool Mastery with a Flick of Your Wrist',
    'long_description': 'None',
    'author': 'JensRoland',
    'author_email': 'mail@jensroland.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://kegstand.dev/sentient-switchblade',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
